import argparse
import json
import os
import dataclasses
import re
from getpass import getpass
from os import environ, path, makedirs
from typing import Optional, List

from winix import WinixAccount, WinixDevice, WinixDeviceStub
from winix.auth import WinixAuthResponse, login, refresh

DEFAULT_CONFIG_PATH = "~/.config/winix/config.json"


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        return (
            dataclasses.asdict(o) if dataclasses.is_dataclass(o) else super().default(o)
        )


class Configuration:
    exists: bool

    cognito: Optional[WinixAuthResponse]
    devices: List[WinixDeviceStub]

    def __init__(self, config_path: str):
        self.config_path = path.expanduser(config_path)
        self._load_from_disk()

    @property
    def device(self) -> WinixDeviceStub:
        return self.devices[0]

    def _load_from_disk(self):
        if path.exists(self.config_path):
            with open(self.config_path) as f:
                js = json.load(f)
                self.exists = True
                self.cognito = (
                    WinixAuthResponse(**js["cognito"])
                    if js.get("cognito") is not None
                    else None
                )
                self.devices = [WinixDeviceStub(**d) for d in js.get("devices", [])]
        else:
            self.exists = False
            self.cognito = None
            self.devices = []

    def save(self):
        makedirs(path.dirname(self.config_path), mode=0o755, exist_ok=True)
        with open(os.open(self.config_path, os.O_CREAT | os.O_WRONLY, 0o755), "w") as f:
            json.dump(
                {"cognito": self.cognito, "devices": self.devices,}, f, cls=JSONEncoder
            )


class Cmd:
    def __init__(self, args: argparse.Namespace, config: Configuration):
        self.args = args
        self.config = config


class LoginCmd(Cmd):
    parser_args = {
        "name": "login",
        "help": "Authenticate Winix account",
    }

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("--username", help="Username (email)", required=False)
        parser.add_argument("--password", help="Password", required=False)
        parser.add_argument(
            "--refresh",
            dest="refresh",
            action="store_true",
            help="Refresh the Winix Cognito token instead of logging in",
        )

    def execute(self):
        if getattr(self.args, "refresh"):
            return self._refresh()
        else:
            return self._login()

    def _login(self):
        print(
            "You need to signup for a Winix account & associate your device in the phone app before using this."
        )
        username = getattr(self.args, "username") or input("Username (email): ")
        password = getattr(self.args, "password") or getpass("Password: ")

        self.config.cognito = login(username, password)
        account = WinixAccount(self.config.cognito.access_token)
        account.register_user(username)
        account.check_access_token()
        self.config.devices = account.get_device_info_list()
        self.config.save()
        print("Ok")

    def _refresh(self):
        self.config.cognito = refresh(
            user_id=self.config.cognito.user_id,
            refresh_token=self.config.cognito.refresh_token,
        )
        WinixAccount(self.config.cognito.access_token).check_access_token()
        self.config.save()
        print("Ok")


class DevicesCmd(Cmd):
    parser_args = {
        "name": "devices",
        "help": "List registered Winix devices",
    }

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument(
            "--expose", action="store_true", help="Expose sensitive Device ID string"
        )

    def execute(self):
        expose = getattr(self.args, "expose", False)
        print(f"{len(self.config.devices)} devices:")

        for i, device in enumerate(self.config.devices):
            hidden_deviceid = "_".join(
                [device.id.split("_")[0], "*" * len(device.id.split("_", 1)[1]),]
            )

            fields = (
                ("Device ID", device.id if expose else hidden_deviceid + " (hidden)"),
                ("Mac", device.mac),
                ("Alias", device.alias),
                ("Location", device.location_code),
            )

            label = " (default)" if i == 0 else ""
            print(f"Device#{i}{label} ".ljust(50, "-"))

            for f, v in fields:
                print(f"{f:>15} : {v}")

            print("")

        print("Missing a device? You might need to run refresh.")


class FanCmd(Cmd):
    parser_args = {
        "name": "fan",
        "help": "Fan speed controls",
    }

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument(
            "level",
            help="Fan level",
            choices=["low", "medium", "high", "turbo", "sleep"],
        )

    def execute(self):
        level = self.args.level
        # TODO(Hunter): Support getting the fan state instead of only being able to set it
        device = WinixDevice(self.config.device.id)
        getattr(device, level)()
        print("ok")


class PowerCmd(Cmd):
    parser_args = {
        "name": "power",
        "help": "Power controls",
    }

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("state", help="Power state", choices=["on", "off"])

    def execute(self):
        state = self.args.state
        device = WinixDevice(self.config.device.id)
        getattr(device, state)()
        print("ok")


class ModeCmd(Cmd):
    parser_args = {
        "name": "mode",
        "help": "Mode controls",
    }

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("state", help="Mode state", choices=["auto", "manual"])

    def execute(self):
        state = self.args.state
        device = WinixDevice(self.config.device.id)
        getattr(device, state)()
        print("ok")


class PlasmawaveCmd(Cmd):
    parser_args = {
        "name": "plasmawave",
        "help": "Plasmawave controls",
    }

    @classmethod
    def add_parser(cls, parser):
        parser.add_argument("state", help="Plasmawave state", choices=["on", "off"])

    def execute(self):
        state = "plasmawave_on" if self.args.state == "on" else "plasmawave_off"
        device = WinixDevice(self.config.device.id)
        getattr(device, state)()
        print("ok")


class RefreshCmd(Cmd):
    parser_args = {
        "name": "refresh",
        "help": "Refresh account device metadata",
    }

    @classmethod
    def add_parser(cls, parser):
        pass

    def execute(self):
        account = WinixAccount(self.config.cognito.access_token)
        self.config.devices = account.get_device_info_list()
        self.config.save()
        print("Ok")


def main():
    parser = argparse.ArgumentParser(description="Winix C545 Air Purifier Control")
    subparsers = parser.add_subparsers(dest="cmd")

    commands = {
        cls.parser_args["name"]: cls
        for cls in (
            LoginCmd,
            RefreshCmd,
            DevicesCmd,
            FanCmd,
            PowerCmd,
            ModeCmd,
            PlasmawaveCmd,
        )
    }

    for cls in commands.values():
        sub = subparsers.add_parser(**cls.parser_args)
        cls.add_parser(sub)

    args = parser.parse_args()
    cmd = args.cmd

    if cmd is None:
        parser.print_help()
        return

    cls = commands[cmd]
    cls(args, config=Configuration("~/.config/winix/config.json")).execute()



import dataclasses
from binascii import crc32
from typing import Optional

import requests


@dataclasses.dataclass
class WinixDeviceStub:
    id: str
    mac: str
    alias: str
    location_code: str
    filter_replace_date: str


class WinixAccount:
    def __init__(self, access_token: str):
        self._uuid: Optional[str] = None
        self.access_token = access_token

    def check_access_token(self):
        """Register the Cognito Token with the Winix backen (again)"""
        from winix import auth

        payload = {
            "cognitoClientSecretKey": auth.COGNITO_CLIENT_SECRET_KEY,
            "accessToken": self.access_token,
            "uuid": self.get_uuid(),
            "osVersion": "26",  # oreo
            "mobileLang": "en",
        }

        resp = requests.post(
            "https://us.mobile.winix-iot.com/checkAccessToken", json=payload
        )

        if resp.status_code != 200:
            raise Exception(
                f"Error while performing RPC checkAccessToken ({resp.status_code}): {resp.text}"
            )

    def get_device_info_list(self):
        resp = requests.post(
            "https://us.mobile.winix-iot.com/getDeviceInfoList",
            json={"accessToken": self.access_token, "uuid": self.get_uuid(),},
        )

        if resp.status_code != 200:
            raise Exception(
                f"Error while performing RPC checkAccessToken ({resp.status_code}): {resp.text}"
            )

        return [
            WinixDeviceStub(
                id=d["deviceId"],
                mac=d["mac"],
                alias=d["deviceAlias"],
                location_code=d["deviceLocCode"],
                filter_replace_date=d["filterReplaceDate"],
            )
            for d in resp.json()["deviceInfoList"]
        ]

    def register_user(self, email: str):
        """Register the logged-in android login/android user uuid with the backend"""
        # Call after getting a cognito token but before check_access_token
        # necessary for the winix backend to recognize the Android "uuid" we send
        # in most API requests
        from winix import auth

        resp = requests.post(
            "https://us.mobile.winix-iot.com/registerUser",
            json={
                "cognitoClientSecretKey": auth.COGNITO_CLIENT_SECRET_KEY,
                "accessToken": self.access_token,
                "uuid": self.get_uuid(),
                "email": email,
                "osType": "android",
                "osVersion": "29",
                "mobileLang": "en",
            },
        )

        if resp.status_code != 200:
            raise Exception(
                f"Error while performing RPC registerUser ({resp.status_code}): {resp.text}"
            )

    def get_uuid(self) -> str:
        # We construct our fake secure Android ID as
        # CRC32("github.com/hfern/winixctl" + userid) + CRC32("HGF" + userid)
        # where userid is the formatted uuid string from cognito

        if self._uuid is None:
            from jose import jwt

            userid_b = jwt.get_unverified_claims(self.access_token)["sub"].encode()
            p1 = crc32(b"github.com/hfern/winixctl" + userid_b)
            p2 = crc32(b"HGF" + userid_b)
            self._uuid = f"{p1:08x}{p2:08x}"

        return self._uuid


class WinixDevice:
    URL = "https://us.api.winix-iot.com/common/control/devices/{deviceid}/A211/{attribute}:{value}"

    K_POWER = "A02"
    V_POWER_STATES = {
        "off": "0",
        "on": "1",
    }

    K_MODE = "A03"
    V_MODE_STATES = {"auto": "01", "manual": "02"}

    K_AIRFLOW = "A04"
    V_AIRFLOW_STATES = {
        "low": "01",
        "medium": "02",
        "high": "03",
        "turbo": "05",
        "sleep": "06",
    }

    K_PLASMAWAVE = "A07"
    V_PLASMAWAVE_STATES = {
        "off": "0",
        "on": "1",
    }

    K_PLASMA = "A07"
    V_PLASMA_STATES = {
        "off": "0",
        "on": "1",
    }

    def __init__(self, id):
        self.id = id

    def off(self):
        self._rpc_attr(self.K_POWER, self.V_POWER_STATES["off"])

    def on(self):
        self._rpc_attr(self.K_POWER, self.V_POWER_STATES["on"])

    def auto(self):
        self._rpc_attr(self.K_MODE, self.V_MODE_STATES["auto"])

    def manual(self):
        self._rpc_attr(self.K_MODE, self.V_MODE_STATES["manual"])

    def plasmawave_off(self):
        self._rpc_attr(self.K_PLASMAWAVE, self.V_PLASMAWAVE_STATES["off"])

    def plasmawave_on(self):
        self._rpc_attr(self.K_PLASMAWAVE, self.V_PLASMAWAVE_STATES["on"])

    def low(self):
        self._rpc_attr(self.K_AIRFLOW, self.V_AIRFLOW_STATES["low"])

    def medium(self):
        self._rpc_attr(self.K_AIRFLOW, self.V_AIRFLOW_STATES["medium"])

    def high(self):
        self._rpc_attr(self.K_AIRFLOW, self.V_AIRFLOW_STATES["high"])

    def turbo(self):
        self._rpc_attr(self.K_AIRFLOW, self.V_AIRFLOW_STATES["turbo"])

    def sleep(self):
        self._rpc_attr(self.K_AIRFLOW, self.V_AIRFLOW_STATES["sleep"])

    def plasma_off(self):
        self._rpc_attr(self.K_PLASMA, self.V_PLASMA_STATES["off"])

    def plasma_on(self):
        self._rpc_attr(self.K_PLASMA, self.V_PLASMA_STATES["on"])

    def _rpc_attr(self, attr: str, value: str):
        requests.get(self.URL.format(deviceid=self.id, attribute=attr, value=value))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="winix",
    version="0.1.1",
    author="Hunter Fernandes",
    author_email="hunter@hfernandes.com",
    description="Programmatically control the Winix C545",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hfern/winix",
    packages=setuptools.find_packages(),
    install_requires=["warrant", "warrant_lite", "requests",],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    entry_points={
        "console_scripts": ["winix = winix.cmd:main", "winixctl = winix.cmd:main",],
    },
)



from dataclasses import dataclass
import boto3

import boto3
from botocore import UNSIGNED
from botocore.client import Config

# Pulled from Winix Home v1.0.8 APK
COGNITO_APP_CLIENT_ID = "14og512b9u20b8vrdm55d8empi"
COGNITO_CLIENT_SECRET_KEY = "k554d4pvgf2n0chbhgtmbe4q0ul4a9flp3pcl6a47ch6rripvvr"
COGNITO_USER_POOL_ID = "us-east-1_Ofd50EosD"
COGNITO_REGION = "us-east-1"


@dataclass
class WinixAuthResponse:
    user_id: str
    access_token: str
    refresh_token: str
    id_token: str


def login(username: str, password: str, **kwargs):
    """Generate fresh credentials"""
    from warrant_lite import WarrantLite
    from jose import jwt

    wl = WarrantLite(
        username=username,
        password=password,
        pool_id=kwargs.get("pool_id", COGNITO_USER_POOL_ID),
        client_id=kwargs.get("client_id", COGNITO_APP_CLIENT_ID),
        client_secret=kwargs.get("client_secret", COGNITO_CLIENT_SECRET_KEY),
        client=_boto_client(kwargs.get("pool_region")),
    )

    resp = wl.authenticate_user()
    return WinixAuthResponse(
        user_id=jwt.get_unverified_claims(resp["AuthenticationResult"]["AccessToken"])[
            "sub"
        ],
        access_token=resp["AuthenticationResult"]["AccessToken"],
        refresh_token=resp["AuthenticationResult"]["RefreshToken"],
        id_token=resp["AuthenticationResult"]["IdToken"],
    )


def refresh(user_id: str, refresh_token: str, **kwargs) -> WinixAuthResponse:
    """Refresh """
    from warrant_lite import WarrantLite

    client_id = kwargs.get("client_id", COGNITO_APP_CLIENT_ID)

    auth_params = {
        "REFRESH_TOKEN": refresh_token,
        "SECRET_HASH": WarrantLite.get_secret_hash(
            username=user_id,
            client_id=client_id,
            client_secret=kwargs.get("client_secret", COGNITO_CLIENT_SECRET_KEY),
        ),
    }

    resp = _boto_client(kwargs.get("pool_region")).initiate_auth(
        ClientId=client_id, AuthFlow="REFRESH_TOKEN", AuthParameters=auth_params,
    )

    return WinixAuthResponse(
        user_id=user_id,
        access_token=resp["AuthenticationResult"]["AccessToken"],
        refresh_token=refresh_token,
        id_token=resp["AuthenticationResult"]["IdToken"],
    )


def _boto_client(region):
    """Get an uncredentialed boto"""
    return boto3.client(
        "cognito-idp",
        config=Config(signature_version=UNSIGNED),
        region_name=region or COGNITO_REGION,
    )
